/*
    rate  --  statistic traffic analyzer
    Copyright (C) 2003 Mateusz 'mteg' Golicz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifdef STREAM_ANALYZER
#include <sys/types.h>
#include <string.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <assert.h>
#include <netdb.h>
#include "lib.h"
#include "util.h"
#include "sta.h"
#include "streams.h"
#include "process.h"

#define REHASH(a, b, h) ((((h) - (a)*d) << 1) + (b))
#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

#define GUESS_AFTER 5

void ftp_ex(sta_stream* s, const unsigned char * data, int len, int end);
void pop3_ex(sta_stream* s, const unsigned char * data, int len, int end);
void smtp_ex(sta_stream* s, const unsigned char * data, int len, int end);
void http_ex(sta_stream* s, const unsigned char * data, int len, int end);

void printstring_info(sta_stream* s, int n, char * buffer);
void pop3_info(sta_stream* s, int n, char * buffer);


void generic_ex(sta_stream* s, const unsigned char * data, int len, int end);
void generic_ds(sta_stream* s);
void generic_info(sta_stream* s, int n, char * buffer);


typedef void (*p_exfunction)(sta_stream*, const unsigned char * data, int len, int end);
typedef void (*p_dsfunction)(sta_stream*);
typedef void (*p_infofunction)(sta_stream*, int, char * buffer);

static p_exfunction exfunctions[] = {
					generic_ex,
					generic_ex,
					generic_ex,
					generic_ex,
					generic_ex,
					generic_ex,
					http_ex,
					ftp_ex,
					generic_ex,
					smtp_ex,
					pop3_ex,
					generic_ex,
					generic_ex,
					generic_ex,
					generic_ex,
					generic_ex};

static p_dsfunction dsfunctions[] = {
					generic_ds,
					generic_ds,
					generic_ds,
					generic_ds,
					generic_ds,
					generic_ds,
					generic_ds,
					generic_ds,
					generic_ds,
					generic_ds,
					generic_ds,
					generic_ds,
					generic_ds,
					generic_ds,
					generic_ds,
					generic_ds};

static p_infofunction infofunctions[] = {
					generic_info,
					generic_info,
					generic_info,
					generic_info,
					generic_info,
					generic_info,
					printstring_info,
					generic_info,
					generic_info,
					printstring_info,
					pop3_info,
					generic_info,
					generic_info,
					generic_info,
					generic_info,
					generic_info};

static char * protonames[] = {
					"Unidentified",
					"Unknown mass transfer protocol",
					"Unknown interactive session protocol",
					"Unknown conversation protocol",
					"ICMP Probe (PING)",
					"ICMP Message stream",
					"HTTP",
					"FTP Control",
					"FTP Data",
					"SMTP",
					"POP3",
					"DNS",
					"SSH",
					"Telnet",
					"TFTP",
					"Ident"};

char * sta_getprotoname(int x)
{
	return(protonames[x]);
}



static inline int rabin_karp(const char *y, int n, const char *x, int m) 
{
    int d, hx, hy, i, j;

    /* Preprocessing */
    /* computes d = 2^(m-1) with
       the left-shift operator */
    for (d = i = 1; i < m; ++i)
       d = (d<<1);

    for (hy = hx = i = 0; i < m; ++i) {
       hx = ((hx<<1) + x[i]);
       hy = ((hy<<1) + y[i]);
    }

    /* Searching */
    j = 0;
    while (j <= n-m) {
       if (hx == hy)
	   	if(memcmp(x, y + j, m) == 0)
          return(j);
		hy = REHASH(y[j], y[j + m], hy);
	    ++j;
	}
	return(-1);
}

static unsigned char * extract(const unsigned char * d, int maxd, int maxs, unsigned char stop1, unsigned char stop2)
{
	int m = min(maxd, maxs);
	unsigned char * str;
	int i;
	
	str = (unsigned char*) malloc(m + 1);
	assert(str);
	for(i = 0; i<m && d[i] != stop1 && d[i] != stop2; i++)
		str[i] = d[i];
	str[i] = 0;
	str = (unsigned char*) realloc(str, i + 1);
	return(str);
}

static unsigned char * extract_after(const unsigned char * d, int maxd, int maxs, char * str, unsigned char stop1, unsigned char stop2)
{
	int pos, l;
	pos = rabin_karp(d, maxd, str, l = strlen(str));
	if(pos == -1) return(NULL);
	return(extract(d + pos + l, maxd - pos - l, maxs, stop1, stop2));
}

#define EX_LIMITLD(s, end, l, d) if(s->protocol_level > l) return; if(s->endpoints[end].dpackets > d) return
#define EX_CLIENT(e) if(e) return
#define EX_SERVER(e) if(!e) return
#define EX_PACKET(s, e, n) if(s->endpoints[e].dpackets != n) return

static int ftppicount = 0;

static struct ftp_portinfo
{
	struct ftp_portinfo * next;
	u_int32_t ip;
	u_int16_t port;
	time_t stamp;
} * portinfos = NULL;

static struct ftp_portinfo * lastpi = NULL;

#define FTPPI_MAX 1000
#define FTPPI_TIMEOUT 150
#define FTPPI_CLEANING_INTERVAL 5

time_t ftp_last_cleaning = 0;

static inline void ftp_cleanpi(void)
{
	time_t now = time(NULL);
	if((now - ftp_last_cleaning) > FTPPI_CLEANING_INTERVAL)
	{
		ftp_last_cleaning = now;
		while(portinfos)
		{
			struct ftp_portinfo * next;
			if((now - portinfos->stamp) < FTPPI_TIMEOUT) break;
			next = portinfos->next;
			free(portinfos);
			portinfos = next;
		}
		if(!portinfos) lastpi = NULL;
	}
}

static inline void ftp_addpi(u_int32_t ip, u_int16_t port)
{
	struct ftp_portinfo * pi = sta_malloc(sizeof(struct ftp_portinfo));
	pi->ip = ip;
	pi->port = port;
	pi->stamp = time(NULL);
	pi->next = NULL;

	if(lastpi)
	{
		lastpi->next = pi;
		lastpi = pi;
	}
	else
	{
		lastpi = pi;
		portinfos = pi;
	}
}

static inline int ftp_isinpi(u_int32_t ip, u_int16_t port)
{
	struct ftp_portinfo *prev = NULL;
	struct ftp_portinfo *pi;
	time_t now = time(NULL);
	
	for(pi = portinfos; pi; prev = pi, pi = pi->next)
	{
		if(pi->ip == ip && pi->port == port && (now - pi->stamp) < FTPPI_TIMEOUT)
		{
			if(prev)
				prev->next = pi->next;
			else
				portinfos = pi->next;

			if(pi == lastpi) lastpi = prev;
			free(pi);
			return(1);
		}
	}
	return(0);
}

void ftp_ex(sta_stream* s, const unsigned char * data, int len, int end)
{
	if(end)
	{

		unsigned char * pasv;
		
		ftp_cleanpi();
		pasv = extract_after(data, len, 250, "Entering Passive Mode ", '\r', '\n');
		if(pasv)
		{
			int i1, i2, i3, i4, p1, p2;
			if(sscanf(pasv, "(%d,%d,%d,%d,%d,%d)", &i1, &i2, &i3, &i4, &p1, &p2) == 6)
			{
				u_int32_t ip = htonl((i1 << 24) | (i2 << 16) | (i3 << 8) | i4);
				u_int16_t port = htons((p1 << 8) | p2);
				ftp_addpi(ip, port);
			}
			free(pasv);
		}
	}
	else
	{
		unsigned char * cwd, * port;
		
		ftp_cleanpi();
		if(!(port = extract_after(data, len, 250, "port ", '\r', '\n')))
			port = extract_after(data, len, 250, "PORT ", '\r', '\n');
		if(port)
		{
			int i1, i2, i3, i4, p1, p2;
			if(sscanf(port, "%d,%d,%d,%d,%d,%d", &i1, &i2, &i3, &i4, &p1, &p2) == 6)
			{
				u_int32_t ip = htonl((i1 << 24) | (i2 << 16) | (i3 << 8) | i4);
				u_int16_t port = htons((p1 << 8) | p2);
				ftp_addpi(ip, port);
			}
			free(port);
		}
	
		if(!(cwd = extract_after(data, len, 250, "cwd ", '\r', '\n')))
			cwd = extract_after(data, len, 250, "CWD ", '\r', '\n');
		if(cwd)
		{
			if(s->protocol_data) free(s->protocol_data);
			s->protocol_data = cwd;
		}
	}
}


void pop3_ex(sta_stream* s, const unsigned char * data, int len, int end)
{
	unsigned char * user;
	EX_CLIENT(end);
	EX_LIMITLD(s, end, 0, 10);
			
	if(!(user = extract_after(data, len, 250, "user ", '\r', '\n')))
		user = extract_after(data, len, 250, "USER ", '\r', '\n');

			
	if((s->protocol_data = user))
		s->protocol_level = 1;
}

void smtp_ex(sta_stream* s, const unsigned char * data, int len, int end)
{
	unsigned char * to, * r;
	
	EX_CLIENT(end);
	EX_LIMITLD(s, end, 1, 10);
	if(!s->protocol_level) /* FROM: */
	{
		unsigned char * from;
		if(!(from = extract_after(data, len, 250, "mail from: ", '\r', '\n')))
			from = extract_after(data, len, 250, "MAIL FROM: ", '\r', '\n');

		if(!from) return;
		s->protocol_data = auto_snprintf(256, "%s => ", from);
		s->protocol_level++;
		free(from);
	}
	if(!(to = extract_after(data, len, 250, "rcpt to: ", '\r', '\n')))
		to = extract_after(data, len, 250, "RCPT TO: ", '\r', '\n');
	if(!to) return;
	r = auto_snprintf(256, "%s%s", s->protocol_data, to);
	free(s->protocol_data);
	free(to);
	s->protocol_data = r;
	s->protocol_level++;
}

void http_ex(sta_stream *s, const unsigned char *data, int len, int end)
{
	unsigned char * host, * req;
	EX_CLIENT(end);
	EX_PACKET(s, end, 2);
	EX_LIMITLD(s, end, 0, 3);
	
	if(len < 10) return;
	req = extract(data, len, 250, '\r', '\n');
	host = extract_after(data, len, 250, "Host: ", '\r', '\n');
	
	if(host)
	{
		s->protocol_data = auto_snprintf(250, "%s:%s", host, req);
		free(req); free(host);
	}
	else
		s->protocol_data = req;
	
	
	s->protocol_level++;
}

void generic_ex(sta_stream *s, const unsigned char *data, int len, int end)
{
}

void generic_info(sta_stream* s, int n, char * buffer)
{
}

void printstring_info(sta_stream* s, int n, char * buffer)
{
	if(s->protocol_data)
		snprintf(buffer, n, "%s", s->protocol_data);
}

void pop3_info(sta_stream* s, int n, char * buffer)
{
	if(s->protocol_data)
		snprintf(buffer, n, "User: %s", s->protocol_data);
}

void generic_ds(sta_stream *s)
{
	free(s->protocol_data);
}

#define THIS_IS(x) { id = (x); }
#define GUESS(x) { return(x); }
inline static int tcp_port_guess(int port)
{
	switch(port)
	{
		case 21:  GUESS(STREAM_FTP_CONTROL);
		case 22:  GUESS(STREAM_SSH);
		case 23:  GUESS(STREAM_TELNET);
		case 25:  GUESS(STREAM_SMTP);
		case 53:  GUESS(STREAM_DOMAIN);
		case 110: GUESS(STREAM_POP3);
		case 113: GUESS(STREAM_AUTH);
	}
	return(0);
}
inline static int udp_port_guess(int port)
{
	switch(port)
	{
		case 53:  GUESS(STREAM_DOMAIN);
		case 69:  GUESS(STREAM_TFTP);
	}
	return(0);
}

int sta_process(sta_stream *s, const unsigned char * content, int len, int endpoint)
{
	int id = 0;
	
	if(s->stream_protocol)
	{
		exfunctions[s->stream_protocol](s, content, len, endpoint);
		return(0);
	}
	
	if((s->properties & SPROP_PROTO_MASK) == SPROP_PROTO_TCP) /* TCP / IP */
	{
		if(endpoint) /* if the sender is not the host that opened the session */
		{
			if(len > 5)
			{
				if(s->endpoints[endpoint].dpackets == 1) /* if it's the first packet after a SYN */
				{
					/* SMTP ? */
					if(strncmp(content, "220 ", 4) == 0 &&
					   ((rabin_karp(content, len, "SMTP", 4) != -1) ||
					    (rabin_karp(content, len, "smtp", 4) != -1)))
						THIS_IS(STREAM_SMTP)
					else if(strncmp(content, "+OK", 3) == 0)
						THIS_IS(STREAM_POP3);
						
				}
				if(!id && s->endpoints[endpoint].dpackets < 4)
				{
					if(rabin_karp(content, len, "220", 3) != -1 &&
						   ((rabin_karp(content, len, "FTP", 3) != -1) ||
					    	(rabin_karp(content, len, "ftp", 3) != -1)))
							THIS_IS(STREAM_FTP_CONTROL)
				}
			}
		}
		else /* if the sender is the host that opened the session */
		{
			if(s->endpoints[endpoint].dpackets == 2) /* if it's the first packet after a SYN */
			{
				/* HTTP traffic ? */
				if(len > 5)
				{
					if(strncmp(content, "GET ", 4) == 0 || 
				   		strncmp(content, "PUT ", 4) == 0 || 
					    strncmp(content, "HEAD ", 5) == 0 || 
					    strncmp(content, "POST ", 5) == 0)
						   THIS_IS(STREAM_HTTP);
				}
			}
			
			if(s->endpoints[endpoint].dpackets == GUESS_AFTER)
			{
				float avpkt1 = 0, avpkt2 = 0;
				float ratio;
				int g;
				
				if(ftp_isinpi(s->endpoints[1].host->ip, s->endpoints[1].port))
				{
					s->protocol_level = 0;
					return(STREAM_FTP_DATA);
				}
				s->protocol_level = LEVEL_GUESS;

				if(!(g = tcp_port_guess(ntohs(s->endpoints[1].port))))
					g = tcp_port_guess(ntohs(s->endpoints[0].port));

				if(g) return(g);
				
				
				if(s->endpoints[0].packets)
					avpkt1 = s->endpoints[0].bytes / s->endpoints[0].packets;
					
				if(s->endpoints[1].packets)
					avpkt2 = s->endpoints[1].bytes / s->endpoints[1].packets;
				
				if(avpkt1 > 0 && avpkt2 > 0)
				{
					if(((avpkt1 > 1200) && (avpkt2 < 400)) ||
					   ((avpkt2 > 1200) && (avpkt1 < 400)))
					   return(STREAM_GENERIC_TRANSFER);

					ratio = avpkt1 / avpkt2;
					if(ratio > 0.8 && ratio < 1.2)
						return(STREAM_GENERIC_CONVERSATION);
					else
						return(STREAM_GENERIC_INTERACTIVE);
				}
			}
		}
	}

	if((s->properties & SPROP_PROTO_MASK) == SPROP_PROTO_TCP) /* UDP / IP */
	{
		if(s->endpoints[endpoint].dpackets == GUESS_AFTER)
		{
			int g;
			if(!(g = udp_port_guess(ntohs(s->endpoints[1].port))))
				g = udp_port_guess(ntohs(s->endpoints[0].port));

			if(g) return(g);
			
		}
	}

	if((s->properties & SPROP_PROTO_MASK) == SPROP_PROTO_ICMP) /* ICMP / IP */
	{
		if(len > 5)
		{
			if(*content == 8)	/* ICMP echo req */
				return(STREAM_ICMP_PROBE);
			return(STREAM_ICMP_MISC);
		}
	}

	if(id) 
	{
		s->protocol_data = NULL;
		s->protocol_level = 0;
		s->stream_protocol = id;
		exfunctions[id](s, content, len, endpoint);
	}
	return(id);
}



void sta_destroyprotodata(sta_stream *s)
{
	if(s->protocol_level == LEVEL_GUESS) return;
	dsfunctions[s->stream_protocol](s);
}

static int servent_opened = 0;
static char qmark_string[] = "??";

static char * serventguess(sta_stream *s)
{
	int p1 = ntohs(s->endpoints[0].port);
	int p2 = ntohs(s->endpoints[1].port);
	struct servent * se;
	char * proto;
	
	if((s->properties & SPROP_PROTO_MASK) == SPROP_PROTO_TCP)
		proto = "tcp";
	else
		proto = "udp";
	
	if(p1 > p2)
	{
		int t;
		t = p1;
		p1 = p2;
		p2 = t;
	}
	if(!servent_opened)
	{
		servent_opened = 1;
		setservent(1);
	}
	if(!(se = getservbyport(htons(p1), proto)))
		se = getservbyport(htons(p2), proto);
	
	if(se)
		return(se->s_name);
	else
		return(qmark_string);
}

char *sta_protoinfo(sta_stream *s)
{
	static char info[300];
	int n;
	if(s->protocol_level == LEVEL_GUESS)
	{
		if(s->stream_protocol)
			snprintf(info, 300, "%s (guessed) (%s?)", protonames[s->stream_protocol], serventguess(s));
		else
			snprintf(info, 300, "Unknown (%s?)", serventguess(s));
	}
	else
	{
		if(s->stream_protocol)
			n = snprintf(info, 300, "%s ", protonames[s->stream_protocol]);
		else
			n = snprintf(info, 300, "%s (%s?) ", protonames[s->stream_protocol], serventguess(s));

		infofunctions[s->stream_protocol](s, 300 - n, info + n);
	}
	return(info);
}
#endif
