/*
    rate  --  statistic traffic analyzer
    Copyright (C) 2003 Mateusz 'mteg' Golicz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef OPEN_BSD
#include <netinet/in_systm.h>
#endif
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/timeb.h>
#include <sys/time.h>
#include <signal.h>
#include <time.h>
#include <pcap.h>
#include "lib.h"
#include "protos.h"
#ifdef HAVE_REGEX
#include <regex.h>
#endif
#ifdef IOCTL_GETWINSIZE
#ifdef SOLARIS_SPARC
#include <sys/termios.h>
#endif
#include <sys/ioctl.h>
#endif
#include "iphash.h"
#include "save.h"
#include "dns.h"
#ifdef STREAM_ANALYZER
#include "sta.h"
#include "streams.h"
#include "streamstat.h"
#endif

static struct timeb starttime;
static struct timeb lasttime;
unsigned long bytes = 0;
unsigned long packets = 0;
unsigned long long totalbytes = 0;
unsigned long long totalpackets = 0;

unsigned char * opt_bpfexpression = NULL;
double opt_delay = 1.0;
int opt_prefix = -1;
unsigned int opt_onsignal = 0;
unsigned int opt_onkeyboard = 0;

int opt_color = 0;
int opt_human = 1;
int opt_counter = -1;
int opt_separator = 0;
int opt_cls = 0;
int opt_dns = 1;
int opt_cols = 80;

#ifdef HAVE_REGEX
char * opt_regex = NULL;
char * opt_extractor = NULL;
char opt_nul_char = '@';
#endif

int opt_caplen = 40;


void (*custom_output)(struct timeb*, int, int, unsigned long long,
					  unsigned long long, unsigned long, unsigned long);
void (*custom_counter)(const unsigned char *, int, int);


void output_callback(int signal)
{
	int time, millitm, miliFromStart, miliFromLast;
	static struct timeb now;

	{
		fd_set fds;
		struct timeval tv;
		if(opt_dns)
		{
			int dnsfd = dns_waitfd();
			double waittime = 1.0;
			
			dns_events(&waittime);
			
			tv.tv_sec = tv.tv_usec = 0;
			FD_ZERO(&fds);
			FD_SET(dnsfd, &fds);
			if(select(dnsfd + 1, &fds, NULL, NULL, &tv) > 0)
				dns_ack();
		}
		if(opt_onkeyboard)
		{
			int quit = 1;
			char c;
	
			tv.tv_sec = tv.tv_usec = 0;
			FD_ZERO(&fds);
			FD_SET(0, &fds);
			while(select(1, &fds, NULL, NULL, &tv) > 0)
			{
				read(0, &c, 1);
				quit = 0;
			}
			tv.tv_sec = tv.tv_usec = 0;
			if(quit) return;
		}
	}
	
	time = lasttime.time;
	millitm = lasttime.millitm;
	ftime(&now);

	miliFromStart = (now.time - starttime.time) * 1000 + now.millitm - starttime.millitm;
	miliFromLast = (now.time - time) * 1000 + now.millitm - millitm;	
	
	if(!opt_onkeyboard)
		if(miliFromLast < opt_delay)
			return;
	
	memcpy(&lasttime, &now, sizeof(struct timeb));
	
	totalbytes += bytes;
	totalpackets += packets;

	if(opt_cls)
		printf("\033[H");

	custom_output(&lasttime, miliFromStart, miliFromLast, totalbytes, totalpackets, bytes, packets);

	bytes = 0;
	packets = 0;

	opt_counter--;
	if(opt_counter == 0) exit(0);
	if(opt_separator)
		printf("- - - - - - - - - - - - - - - - - - -\n");
	if(opt_cls)
	{
		printf("\033[J");
		fflush(NULL);
	}
	
}

void counter_callback(unsigned char *userdata,const struct pcap_pkthdr *hd,const unsigned char *data)
{
#ifdef HAVE_REGEX
	if(opt_regex)
		if(!check_regex(data + opt_prefix, hd->caplen - opt_prefix, opt_nul_char)) return;
#endif
	bytes += hd->len - opt_prefix;
	packets += 1;

	custom_counter(data + opt_prefix, hd->caplen - opt_prefix, hd->len);
}


void help(unsigned char * name)
{
	printf("rate 0.9 - Mateusz 'mteg' Golicz <mtg@elsat.net.pl>, 2003\n", name);
	printf("usage: %s [-h | -?]\n", name);
	printf("       %s [mode select option] [-h | -?]\n", name);
	printf("       %s -L <name>\n", name);
	printf("       %s [filtering/generic options] [mode select option] [mode options]\n", name);
	printf("\nMODES COMPILED INTO THIS BINARY OF RATE:\n");
	printf(" -R  Rate estimator\n");
	printf("     Use this mode to estimate bandwidth utilized by packets matching\n");
	printf("     given filtering options. This is the default mode.\n");
	printf("\n");
	printf(" -A  Bandwidth abusers\n");
	printf("     This mode is for determining IPs of hosts that consume the highest\n");
	printf("     amount of available bandwidth.\n");
	printf("\n");
#ifdef STREAM_ANALYZER
	printf(" -T  Stream analyzer\n");
	printf("     Using this mode you can have a deeper look on TCP connections and\n");
	printf("     ICMP and UDP streams detected on an interface.\n");
	printf("\n");
#endif
#ifdef HAVE_REGEX
	printf(" -E  Regular expression extractor\n");
	printf("     Use this mode to extract strings from packets.\n");
	printf("\n");
#endif
	printf("\nGENERIC OPTIONS:\n");
	printf("  -h  -?    Show this help\n");
	printf("  -r <t>    Print reports every <t> seconds (default: 1)\n");
	printf("  -g        Dump reports on SIGUSR1, ignore timing.\n");
	printf("  -k        Dump reports on newline on stdin, ignore timing.\n");
	printf("  -q <r>    Quit after printing r reports.\n");
	printf("  -l        Make stdout line-buffered.\n");
	printf("  -p <pref> Datalink layer header size (gets substracted from each packet size, default: 14)\n");
	printf("  -s <b>    Capture l bytes (default: 40)\n");
	printf("  -i <int>  Bind to interface <int> - default eth0\n");
	printf("  -n        Numeric IPs. Don't do reverse DNS lookups.\n");
	printf("  -c        Use colors (ANSI-compatible) whenever possible.\n");
	printf("  -v        Print exact values, do not use SI prefixes.\n");
	printf("  -e        Output a separator after every report to improve readability.\n");
	printf("  -w        Clear the screen before dumping a report.\n");
	printf("  -S <name> Save operation mode (saves all specified command line options).\n");
	printf("  -L <name> Recall operation mode (recalls previously saved option set).\n");
	printf("\nFILTERING OPTIONS:\n");
	printf("  -f <bpf>  BPF filter expression to use\n");
#ifdef HAVE_REGEX
	printf("  -x <rege> Match this regex in packet (increases capture length to at least 1500b)\n");
	printf("  -0 n      Replace nul character with this before doing regex matches (default: '@')\n");
#endif	
}




int main(int argc, char ** argv)
{
	pcap_t *handle;
	unsigned char * interface = "eth0";

	struct bpf_program bpf;
	char ebuff[PCAP_ERRBUF_SIZE];

	int recalled = 0, c;

	struct itimerval itv;

		
	if(argc > 1)
	{
		if(strcmp(argv[1], "-L") == 0)
		{
			if(argc < 3)
			{
				fprintf(stderr, "Specify option set name!\n");
				exit(1);
			}
			else if(argc > 3)
			{
				fprintf(stderr, "No other options allowed after -L\n");
				exit(1);
			}
			recall_args(argv[2], &argv, &argc);
			recalled = 1;
		}
	}
	
	while((c = getopt(argc, argv, "?hr:q:f:p:s:i:S:x:0:glkcvwenRATE")) != EOF)
	{
		if(c == 'R' || c == 'A' || c == 'T' || c == 'E')
			break;

		switch(c)
		{
#if HAVE_REGEX
			case '0':
				opt_nul_char = *optarg;
				break;
			case 'x':
				if(opt_caplen < 1500) opt_caplen = 1500;
				opt_regex = strdup(optarg);
				break;
#endif
			case 'g': opt_onsignal = 1; break;
			case 'k': opt_onkeyboard = 1; break;
			case 'c': opt_color = 1; break;
			case 'v': opt_human = 0; break;
			case 'e': opt_separator = 1; break;
			case 'w': opt_cls = 1; break;
			case 'n': opt_dns = 0; break;
			case 'l': setlinebuf(stdout); break;
			case 'q': opt_counter = atoi(optarg); break;
			case 'p': opt_prefix = atoi(optarg); break;
			case 's': opt_caplen = atoi(optarg); break;
			case 'h':
			case '?':
				help(argv[0]);
				exit(0);
			case 'i':
				interface = strdup(optarg);
				break;
			case 'f':
				opt_bpfexpression = strdup(optarg);
				break;
			case 'r':
				opt_delay = atof(optarg) * 1000;
				if(opt_delay < 0.1) opt_delay = 0.1;
				break;
			case 'S':
				if(!recalled) save_args(optarg, argv, argc);
				break;
		}
	}

	if((handle = pcap_open_live(interface, 2048, -1, 50, ebuff))==NULL)
	{
		fprintf(stderr, "pcap_open_live(): %s\n", ebuff);
		return(0);
 	}

#ifdef AUTOSENSE_DLH
	if(opt_prefix == -1)
		opt_prefix = llheader(pcap_datalink(handle));
#else
	if(opt_prefix == -1)
		opt_prefix = 14;
#endif

	switch(c)
	{
		case 'R':
			r_estimator_setup(argc, argv, &custom_output, &custom_counter);
			break;
		case 'A':
			r_abusers_setup(argc, argv, &custom_output, &custom_counter);
			break;
#ifdef STREAM_ANALYZER
		case 'T':
			r_stream_setup(argc, argv, &custom_output, &custom_counter);
			break;
#endif
#ifdef HAVE_REGEX
		case 'E':
			r_extractor_setup(argc, argv, &custom_output, &custom_counter);
			opt_separator = 0;
			break;
#endif
		default:
			fprintf(stderr, "Unknown operation mode. Consult %s -h.\n", argv[0]);
			exit(1);
			break;
	}

	if(opt_onkeyboard && opt_onsignal)
	{
		fprintf(stderr, "Conflicting options: -k and -g.\n");
		exit(1);
	}

	signal(SIGPIPE, SIG_IGN);
	signal(SIGPROF, SIG_IGN);

	if(opt_onsignal)
	{
		signal(SIGUSR1, output_callback);
		signal(SIGALRM, SIG_IGN);
	}
	else
		signal(SIGALRM, output_callback);
	
	if(argv[optind])
	{
		fprintf(stderr, "Unrecognized command line parameters (starting from: %s)...\n", argv[optind]);
		exit(1);
	}

#ifdef HAVE_REGEX	
	if(opt_regex)
		prepare_regex(opt_regex, opt_caplen);
#endif


	if(opt_bpfexpression)
	{
		if(pcap_compile(handle, &bpf, opt_bpfexpression, 1, 0xffffffff) == -1)
		{
			fprintf(stderr, "pcap_compile(): %s\n", pcap_geterr(handle));
			return(1);
		}
		if(pcap_setfilter(handle, &bpf) == -1)
		{
			fprintf(stderr, "pcap_compile(): %s\n", pcap_geterr(handle));
			return(1);
		}
	}

	
	ftime(&starttime);
	ftime(&lasttime);

	itv.it_interval.tv_sec = 0;
	itv.it_interval.tv_usec = 200000;
	itv.it_value.tv_sec = 0;
	itv.it_value.tv_usec = 200000;
	setitimer(ITIMER_REAL, &itv, NULL);

	itv.it_interval.tv_sec = PROFILER_STARTUP;
	itv.it_interval.tv_usec = 0;
	itv.it_value.tv_sec = PROFILER_STARTUP;
	itv.it_value.tv_usec = 0;
	setitimer(ITIMER_PROF, &itv, NULL);

	if(opt_dns) dns_open();
	if(opt_cls) printf("\033[H\033[J");

#ifdef IOCTL_GETWINSIZE
	{
		struct winsize ws;
		if(ioctl(0, TIOCGWINSZ, &ws) == 0)
		{
			opt_cols = ws.ws_col;
			if(opt_cols < 80) opt_cols = 80;
		}
	}
#endif


	pcap_loop(handle, -1, counter_callback, NULL);

	pcap_close(handle);
}

